@@ -66,6 +66,7 @@ enum {
 	LOOK_SIGNAL,
 	UNLOOK_SIGNAL,
 	WIN_SIGNAL,
+	HINT_SIGNAL,
 	LAST_SIGNAL
 };
 
@@ -109,30 +110,28 @@ static inline gint cell_idx(GtkMineField *mfield, guint x, guint y)
 
 static void _setup_sign (sign *signp, const char *file, guint minesize)
 {
-	GdkPixbuf *scaledimage;
-
 	if (signp->pixbuf == NULL) {
 		GError *error = NULL;
 		signp->pixbuf = gdk_pixbuf_new_from_file (file, &error);
 		if (signp->pixbuf == NULL)
 			g_error (error->message);
-	}	
+	} else {
+		g_object_unref (signp->scaledpixbuf);
+	}
 
-        scaledimage = gdk_pixbuf_scale_simple (signp->pixbuf, minesize - 2, minesize - 2,
-                                               GDK_INTERP_BILINEAR);
-        
-	gdk_pixbuf_render_pixmap_and_mask (scaledimage, &signp->pixmap,
-					   &signp->mask, 127);
-
-        g_object_unref (scaledimage);
-	gdk_drawable_get_size(signp->pixmap, &(signp->width), &(signp->height));
+        signp->scaledpixbuf = gdk_pixbuf_scale_simple (signp->pixbuf, 
+						       minesize - 2, 
+						       minesize - 2,
+						       GDK_INTERP_BILINEAR);
+	signp->width = gdk_pixbuf_get_width (signp->scaledpixbuf);
+	signp->height = gdk_pixbuf_get_height (signp->scaledpixbuf);
 }
 
 static void gtk_minefield_setup_signs(GtkMineField *mfield)
 {
-        _setup_sign(&mfield->flag, DATADIR"/pixmaps/gnomine/flag.png", mfield->minesize);
-        _setup_sign(&mfield->mine, DATADIR"/pixmaps/gnomine/mine.png", mfield->minesize);
-        _setup_sign(&mfield->question, DATADIR"/pixmaps/gnomine/flag-question.png", mfield->minesize);
+        _setup_sign(&mfield->flag, DATADIR"/pixmaps/gnomine/flag.svg", mfield->minesize);
+        _setup_sign(&mfield->mine, DATADIR"/pixmaps/gnomine/mine.svg", mfield->minesize);
+        _setup_sign(&mfield->question, DATADIR"/pixmaps/gnomine/flag-question.svg", mfield->minesize);
 }
 
 static void
@@ -156,6 +155,8 @@ gtk_minefield_setup_numbers (GtkMineField *mfield)
 		PangoAttrList *alist;
 		PangoAttribute *attr;
 		PangoFontDescription *font_desc;
+		PangoRectangle extent;
+		guint64 font_size;
 
 		/* free an existing layout ... */
 		if (mfield->numstr[i].layout)
@@ -169,19 +170,6 @@ gtk_minefield_setup_numbers (GtkMineField *mfield)
 		/* set attributes for the layout */
 		alist = pango_attr_list_new();
 
-		/* do the font */
-		font_desc = pango_font_description_new();
-		pango_font_description_set_family(font_desc, "Mono 8");
-		pango_font_description_set_size(font_desc,
-						pixel_sz * PANGO_SCALE * 0.7);
-		pango_font_description_set_weight(font_desc,PANGO_WEIGHT_BOLD);
-		attr = pango_attr_font_desc_new(font_desc);
-		pango_font_description_free(font_desc);
-
-		attr->start_index = 0;
-		attr->end_index = G_MAXUINT;
-		pango_attr_list_insert(alist, attr);
-
 		/* colour */
 		attr = pango_attr_foreground_new(num_colors[i][0],
 						 num_colors[i][1],
@@ -190,17 +178,46 @@ gtk_minefield_setup_numbers (GtkMineField *mfield)
 		attr->end_index = G_MAXUINT;
 		pango_attr_list_insert(alist, attr);
 
-		pango_layout_set_attributes(layout, alist);
-		pango_attr_list_unref(alist);
+		/* do the font */
+		font_desc = pango_font_description_new();
+		pango_font_description_set_family(font_desc, "Mono 8");
+		/* Guess at the correct font size. */
+		font_size = pixel_sz * PANGO_SCALE;
+		pango_font_description_set_size(font_desc, font_size);
+		pango_font_description_set_weight(font_desc,PANGO_WEIGHT_BOLD);
+		attr = pango_attr_font_desc_new(font_desc);
 
-		pango_layout_set_width(layout, minesize);
+		attr->start_index = 0;
+		attr->end_index = G_MAXUINT;
+		pango_attr_list_insert(alist, attr);
+
+		pango_layout_set_attributes(layout, alist);
 		pango_layout_set_alignment(layout, PANGO_ALIGN_CENTER);
 
+		/* Now correct the size to fit. */
+		pango_layout_get_extents (layout, &extent, NULL);
+		font_size = font_size*0.7*pixel_sz*PANGO_SCALE/extent.height;
+		pango_font_description_set_size(font_desc, font_size);
+		attr = pango_attr_font_desc_new(font_desc);
+		attr->start_index = 0;
+		attr->end_index = G_MAXUINT;
+		pango_attr_list_insert(alist, attr);		
+		pango_layout_set_attributes(layout, alist);
+
+		pango_font_description_free(font_desc);
+		pango_attr_list_unref(alist);
+
 		mfield->numstr[i].layout = layout;
 
-		mfield->numstr[i].dx = minesize / 2;
-		/* Hack to fix the number being too far down */
-		mfield->numstr[i].dy = (minesize - pixel_sz) / 2 - 4;
+		pango_layout_get_extents (layout, NULL, &extent);
+
+		/* The +1 is necessary since these coordinates are
+		 * with respect to minesize, not pixel_sz (the
+		 * difference is 2). */
+		mfield->numstr[i].dx = (pixel_sz - extent.width/PANGO_SCALE)/2 
+			+ 1;
+		mfield->numstr[i].dy = (pixel_sz - extent.height/PANGO_SCALE)/2
+			+ 1;
 	}
 }
 
@@ -271,6 +288,15 @@ static void gtk_minefield_size_allocate (GtkWidget     *widget,
                 xofs = allocation->x + (allocation->width - width)/2;
                 yofs = allocation->y + (allocation->height - height)/2;
                 
+		if (!mfield->thick_line)
+			mfield->thick_line = gdk_gc_new (widget->window);
+		gdk_gc_copy (mfield->thick_line, widget->style->black_gc);
+		gdk_gc_set_line_attributes (mfield->thick_line, 
+					    MAX(1, 0.1*minesize),
+					    GDK_LINE_SOLID,
+					    GDK_CAP_ROUND,
+					    GDK_JOIN_ROUND);
+
 		gdk_window_move_resize (widget->window, xofs, yofs,
                                         width, height);
 	}
@@ -298,6 +324,8 @@ static void gtk_mine_draw(GtkMineField *mfield, guint x, guint y)
         GtkWidget *widget = GTK_WIDGET(mfield);
 	GdkRectangle rect;
 	
+	g_return_if_fail (c != -1);
+
         /* This gives us a dotted line to increase the contrast between
          * buttons and the "sea". */
         if (stipple == NULL) {
@@ -311,13 +339,9 @@ static void gtk_mine_draw(GtkMineField *mfield, guint x, guint y)
                 
 	minesize = mfield->minesize;
 	
-	if (mfield->lose || mfield->win) {
-		noshadow = mfield->mines[c].shown;
-	} else {
-		noshadow = mfield->mines[c].shown;
-	}
+	noshadow = mfield->mines[c].shown;
 
-	clicked = mfield->mines[c].down;
+	clicked = mfield->mines[c].down; 
 
 	/* gtk_paint_box needs a clipping rectangle. */
 	rect.x = x*minesize;
@@ -327,14 +351,14 @@ static void gtk_mine_draw(GtkMineField *mfield, guint x, guint y)
 
 	if (noshadow) { /* draw grid on ocean floor */
 		gtk_paint_box (widget->style,
-				    widget->window,
-				    GTK_STATE_NORMAL,
-				    clicked ? GTK_SHADOW_IN : GTK_SHADOW_NONE,
-				    &rect,
-				    widget,
-				    "button",
-				    x*minesize, y*minesize,
-				    minesize, minesize);
+			       widget->window,
+			       clicked ? GTK_STATE_ACTIVE : GTK_STATE_NORMAL,
+			       GTK_SHADOW_IN,
+			       &rect,
+			       widget,
+			       "button",
+			       x*minesize, y*minesize,
+			       minesize, minesize);
 		if (y == 0) {	/* top row only */
 			gdk_draw_line(widget->window,	/* top */
 			              dots,
@@ -367,7 +391,7 @@ static void gtk_mine_draw(GtkMineField *mfield, guint x, guint y)
 	} else {	/* draw shadow around possible mine location */
 		gtk_paint_box (widget->style,
 			       widget->window,
-			       GTK_STATE_SELECTED,
+			       clicked ? GTK_STATE_ACTIVE : GTK_STATE_SELECTED,
 			       clicked ? GTK_SHADOW_IN : GTK_SHADOW_OUT,
 			       &rect,
 			       widget,
@@ -387,79 +411,40 @@ static void gtk_mine_draw(GtkMineField *mfield, guint x, guint y)
 		}
 
 	} else if (mfield->mines[c].marked == MINE_QUESTION) {
-		gdk_gc_set_clip_mask(widget->style->black_gc,
-		                     mfield->question.mask);
-		gdk_gc_set_clip_origin(widget->style->black_gc,
-		                       x * minesize + (minesize - mfield->question.width) / 2, 
-		                       y * minesize + (minesize - mfield->question.height) / 2);
-		gdk_draw_drawable (widget->window,
-		                 widget->style->black_gc,
-		                 mfield->question.pixmap, 0, 0,
-		                 x * minesize + (minesize - mfield->question.width) / 2, 
-		                 y * minesize + (minesize - mfield->question.height) / 2,
-		                 -1, -1);
-		gdk_gc_set_clip_mask(widget->style->black_gc, NULL);
-	  
-	} else if (mfield->mines[c].marked == MINE_MARKED) {
-		gdk_gc_set_clip_mask(widget->style->black_gc,
-					     mfield->flag.mask);
-		gdk_gc_set_clip_origin(widget->style->black_gc,
-					       x * minesize + (minesize - mfield->flag.width) / 2, 
-					       y * minesize + (minesize - mfield->flag.height) / 2);
-		gdk_draw_drawable (widget->window,
-				 widget->style->black_gc,
-		                 mfield->flag.pixmap, 0, 0,
-		                 x * minesize + (minesize - mfield->flag.width) / 2, 
+		gdk_draw_pixbuf (widget->window, NULL,
+				 mfield->question.scaledpixbuf, 0, 0, 
+				 x * minesize + (minesize - mfield->flag.width) / 2, 
 		                 y * minesize + (minesize - mfield->flag.height) / 2,
-		                 -1, -1);
-	        gdk_gc_set_clip_mask(widget->style->black_gc, NULL);
-	  
+				 mfield->flag.width, mfield->flag.height,
+		                 GDK_RGB_DITHER_NORMAL, 0, 0);
+	} else if (mfield->mines[c].marked == MINE_MARKED) {
+		gdk_draw_pixbuf (widget->window, NULL,
+				 mfield->flag.scaledpixbuf, 0, 0, 
+				 x * minesize + (minesize - mfield->flag.width) / 2, 
+		                 y * minesize + (minesize - mfield->flag.height) / 2,
+				 mfield->flag.width, mfield->flag.height,
+		                 GDK_RGB_DITHER_NORMAL, 0, 0);
+
 		if (mfield->lose && mfield->mines[c].mined != 1) {
-			gdk_draw_line(widget->window,
-				      widget->style->black_gc,
-				      x*minesize+2,
-				      y*minesize+3,
-				      x*minesize+minesize-4,
-				      y*minesize+minesize-3);
-			gdk_draw_line(widget->window,
-				      widget->style->black_gc,
-				      x*minesize+3,
-				      y*minesize+2,
-				      x*minesize+minesize-3,
-				      y*minesize+minesize-4);
-			gdk_draw_line(widget->window,
-				      widget->style->black_gc,
-				      x*minesize+2,
-				      y*minesize+minesize-4,
-				      x*minesize+minesize-4,
-				      y*minesize+2);
-			gdk_draw_line(widget->window,
-				      widget->style->black_gc,
-				      x*minesize+3,
-				      y*minesize+minesize-3,
-				      x*minesize+minesize-3,
-				      y*minesize+3);
-		}
+			int x1 = x*minesize + 0.1*minesize;
+			int y1 = y*minesize + 0.1*minesize;
+			int x2 = x*minesize + 0.9*minesize;
+			int y2 = y*minesize + 0.9*minesize;
 
+			gdk_draw_line (widget->window,
+				       mfield->thick_line,
+				       x1, y1, x2, y2);
+			gdk_draw_line (widget->window,
+				       mfield->thick_line,
+				       x1, y2, x2, y1);
+		}
 	} else if (mfield->lose && mfield->mines[c].mined) {
-		if (mfield->mine.mask) {
-			gdk_gc_set_clip_mask(widget->style->black_gc,
-					     mfield->mine.mask);
-			gdk_gc_set_clip_origin(widget->style->black_gc,
-					       x*minesize + (minesize - mfield->mine.width) / 2, 
-					       y*minesize + (minesize - mfield->mine.height) / 2);
-		}
-
-	        gdk_draw_drawable (widget->window,
-				 widget->style->black_gc,
-		                 mfield->mine.pixmap, 0, 0,
-		                 x * minesize + (minesize - mfield->mine.width) / 2,
-		                 y * minesize + (minesize - mfield->mine.height) / 2,
-		                 -1, -1);
-		
-		if (mfield->flag.mask) {
-			gdk_gc_set_clip_mask(widget->style->black_gc, NULL);
-		}
+		gdk_draw_pixbuf (widget->window, NULL,
+				 mfield->mine.scaledpixbuf, 0, 0, 
+				 x * minesize + (minesize - mfield->flag.width) / 2, 
+		                 y * minesize + (minesize - mfield->flag.height) / 2,
+				 mfield->flag.width, mfield->flag.height,
+		                 GDK_RGB_DITHER_NORMAL, 0, 0);
 	}
 }
 
@@ -643,7 +628,9 @@ static void gtk_minefield_randomize (GtkMineField *mfield, int curloc)
 
 static void gtk_minefield_show(GtkMineField *mfield, guint x, guint y)
 {
-	int c = x + mfield->xsize * y;
+	int c = cell_idx (mfield, x, y);
+
+	g_return_if_fail (c != -1);
 
 	/* make sure first click isn't on a mine */
 	if (!mfield->in_play) {
@@ -671,12 +658,23 @@ static void gtk_minefield_toggle_mark(GtkMineField *mfield, guint x, guint y)
 {
         int c = cell_idx(mfield, x, y);
 
+	g_return_if_fail (c != -1);
+
 	if (mfield->mines[c].shown != 0) { /* nothing to toggle */
 		return;
 	}
 
 	switch (mfield->mines[c].marked) {
 	case MINE_NOMARK:
+		/* If we've used all the flags don't plant any more,
+		 * this should be an indication to the player that they
+		 * have made a mistake. */
+		if (mfield->flag_count == mfield->mcount) {
+			if (mfield->use_question_marks) {
+				mfield->mines[c].marked = MINE_QUESTION;
+			}
+			break;
+		}
 		mfield->mines[c].marked = MINE_MARKED;
 		mfield->flag_count++;
 		break;
@@ -699,9 +697,6 @@ static void gtk_minefield_toggle_mark(GtkMineField *mfield, guint x, guint y)
 		g_signal_emit(GTK_OBJECT(mfield),
 				minefield_signals[MARKS_CHANGED_SIGNAL],
 				0, NULL);
-	if (mfield->shown == mfield->xsize * mfield->ysize-mfield->mcount) {
-                        gtk_minefield_win(mfield);
-		}
 }
 
 static inline void gtk_minefield_multi_press(GtkMineField *mfield,
@@ -752,7 +747,7 @@ static void gtk_minefield_multi_release (GtkMineField *mfield, guint x, guint y,
                         continue;
                 if (mfield->mines[c2].down) {
                         mfield->mines[c2].down = 0;
-                        if (really) {
+                        if (really && (mfield->mines[c2].shown == 0)) {
 				mfield->mines[c2].shown = 1;
 				mfield->shown++;
                                 if (mfield->mines[c2].mined == 1) {
@@ -797,7 +792,8 @@ static gint gtk_minefield_motion_notify(GtkWidget *widget, GdkEventMotion *event
                 if (x < 0 || y < 0 || x > mfield->xsize-1 || y > mfield->ysize-1)
                         return 0;
 
-                c = x+y*(mfield->xsize);
+                c = cell_idx (mfield, x, y);
+		g_assert (c != -1);
 
                 if (c != mfield->cdown) {
                         mfield->mines[mfield->cdown].down = 0;
@@ -841,7 +837,8 @@ static gint gtk_minefield_button_press(GtkWidget *widget, GdkEventButton *event)
         if (event->button <= 3 && !mfield->bdown[1]) {
                 x = event->x/minesize;
                 y = event->y/minesize;
-                c = x+y*(mfield->xsize);
+                c = cell_idx (mfield, x, y);
+		if (c == -1) return FALSE;
                 if (!mfield->bdown[0] && !mfield->bdown[1] && !mfield->bdown[2]) {
                         mfield->cdownx = x;
                         mfield->cdowny = y;
@@ -978,6 +975,16 @@ static void gtk_minefield_class_init (GtkMineFieldClass *class)
 			       g_cclosure_marshal_VOID__VOID,
 			       G_TYPE_NONE,
 			       0);
+	minefield_signals[HINT_SIGNAL] =
+		g_signal_new("hint-used",
+			       G_OBJECT_CLASS_TYPE (object_class),
+			       G_SIGNAL_RUN_FIRST,
+			       G_STRUCT_OFFSET(GtkMineFieldClass, hint_used),
+			       NULL, NULL,
+			       g_cclosure_marshal_VOID__VOID,
+			       G_TYPE_NONE,
+			       0);
+
 }
 
 static void gtk_minefield_init (GtkMineField *mfield)
@@ -992,6 +999,7 @@ static void gtk_minefield_init (GtkMineField *mfield)
         mfield->mine.pixbuf = NULL;
 	mfield->question.pixbuf = NULL;
 	mfield->grand = g_rand_new ();
+	mfield->thick_line = NULL;
 }
 
 void gtk_minefield_set_size(GtkMineField *mfield, guint xsize, guint ysize)
@@ -1169,7 +1177,13 @@ gint gtk_minefield_hint (GtkMineField *mfield)
 	} else return MINEFIELD_HINT_ALL_MINES;
 
 	x = i % mfield->xsize;
-	y = i / mfield->ysize;
+	y = i / mfield->xsize;
+
+	/* Makes sure that the program knows about the successful
+	 * hint before a possible win. */
+        g_signal_emit(GTK_OBJECT(mfield),
+                        minefield_signals[HINT_SIGNAL],
+			0, NULL);
 
 	gtk_minefield_show (mfield, x, y);
 	gtk_mine_draw (mfield, x, y);
