@@ -823,10 +823,6 @@ GnomeUIInfo gamemenu [] = {
 
 	 GNOMEUIINFO_MENU_HINT_ITEM(hint_callback, NULL),
 
-         {GNOME_APP_UI_ITEM, N_("Shu_ffle tiles"), N_("Shuffle tiles"),
-		 shuffle_tiles_callback, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
-		 NULL, 0, 0, NULL},
-
 	 GNOMEUIINFO_SEPARATOR,
 
          GNOMEUIINFO_MENU_SCORES_ITEM(scores_callback, NULL),
@@ -887,10 +883,6 @@ GnomeUIInfo toolbar_uiinfo [] = {
 		NULL, NULL, GNOME_APP_PIXMAP_STOCK, GTK_STOCK_HELP, GDK_H,
 		GDK_CONTROL_MASK, NULL},
 
-	{GNOME_APP_UI_ITEM, N_("Shuffle"), N_("Shuffle tiles"),
-		shuffle_tiles_callback, NULL, NULL,
-		GNOME_APP_PIXMAP_STOCK, GTK_STOCK_EXECUTE, 0, 0, NULL},
-
 	{GNOME_APP_UI_ENDOFINFO}
 };
 
@@ -913,11 +905,6 @@ static void set_menus_sensitive (void)
 	gtk_widget_set_sensitive (gamemenu[7].widget, state);
 	gtk_widget_set_sensitive (toolbar_uiinfo[6].widget, state);
 
-	/* Shuffle */
-	state = game_over != GAME_WON;
-	gtk_widget_set_sensitive (gamemenu[5].widget, state);
-	gtk_widget_set_sensitive (toolbar_uiinfo[6].widget, state);
-
 	/* Restart */
 	state = sequence_number != 1;
 	gtk_widget_set_sensitive (gamemenu[2].widget, state);
@@ -1081,6 +1068,7 @@ mapset_changed_cb (GConfClient *client,
 {
 	GtkWidget *dialog;
 	char *mapset_tmp;
+	gint response;
 
 	mapset_tmp = gconf_client_get_string (conf_client,
 			"/apps/mahjongg/mapset",
@@ -1092,15 +1080,24 @@ mapset_changed_cb (GConfClient *client,
 	} 
 
 	new_map = TRUE;
-	
-	dialog = gtk_message_dialog_new (
+
+	dialog = gtk_message_dialog_new_with_markup (
 		GTK_WINDOW (window),
 		GTK_DIALOG_MODAL,
-		GTK_MESSAGE_INFO,
-		GTK_BUTTONS_OK,
-		_("This new mapset will take effect when you start "
-		  "a new game, or when Mahjongg is restarted."));
-	gtk_dialog_run (GTK_DIALOG (dialog));
+		GTK_MESSAGE_QUESTION,
+		GTK_BUTTONS_NONE,
+		"<b>%s</b>\n\n%s", 
+		_("Do you want to finish the current game or start playing with the new map immediately ?"),
+		_("If you choose to finish with the old map then the next game will use the new map."));
+	gtk_dialog_add_buttons (GTK_DIALOG (dialog),
+				_("_Finish"), GTK_RESPONSE_REJECT,
+				GTK_STOCK_NEW, GTK_RESPONSE_ACCEPT,
+				NULL);
+	gtk_dialog_set_default_response (GTK_DIALOG (dialog),
+					 GTK_RESPONSE_ACCEPT);
+	response = gtk_dialog_run (GTK_DIALOG (dialog));
+	if (response == GTK_RESPONSE_ACCEPT)
+		new_game (FALSE);
 	gtk_widget_destroy (dialog);
 }
 
@@ -1389,18 +1386,19 @@ check_free (void)
 						     GTK_BUTTONS_NONE,
 						     (_("There are no more moves.")));
 			gtk_dialog_add_buttons (GTK_DIALOG (mb),
-						_("Shuffle"),
+						GTK_STOCK_UNDO,
 						GTK_RESPONSE_REJECT,
-						GTK_STOCK_OK,
+						_("Shuffle"),
 						GTK_RESPONSE_ACCEPT,
 						NULL);
 			gtk_dialog_set_default_response (GTK_DIALOG (mb),
 							 GTK_RESPONSE_ACCEPT);
-			if (gtk_dialog_run (GTK_DIALOG (mb)) == GTK_RESPONSE_REJECT)
+			if (gtk_dialog_run (GTK_DIALOG (mb)) == GTK_RESPONSE_ACCEPT)
 				shuffle_tiles_callback (NULL, NULL);
+			else
+				undo_tile_callback (NULL, NULL);
 			gtk_widget_destroy (mb);
                 }
-                game_over = GAME_LOST;
  	} 
 }
 
